@@ -106,7 +106,7 @@ UriStrspnSet uri_strspn_sets[] = {
 #define URI_DELIMITER_IPV6_SET (uri_strspn_sets + 3)
 
 #define BV_SET(bv, idx) (bv)[((guchar)(idx))>>3] |= (1 << ( (idx) & 7) )
-#define BV_IS_SET(bv, idx) ((bv)[(idx)>>3] & (1 << ( (idx) & 7)))
+#define BV_IS_SET(bv, idx) ((bv)[((guchar)(idx))>>3] & (1 << ( (idx) & 7)))
 
 static const char *
 uri_strspn_to(const char *str, UriStrspnSet *set, const char *path_end)
@@ -796,7 +796,7 @@ make_full_uri_from_relative (const char *base_uri, const char *uri)
 			separator = strrchr (mutable_base_uri, '/');
 			if (separator) {
 				/* Make sure we don't eat a domain part */
-				gchar *tmp = (gchar*)((int)separator - 1);
+				char *tmp = separator - 1;
 				if ((separator != mutable_base_uri) && (*tmp != '/')) {
 					*separator = '\0';
 				}
@@ -1175,9 +1175,11 @@ gnome_vfs_uri_to_string (const GnomeVFSURI *uri,
 	}
 
 	if (uri->parent != NULL) {
+		gchar *uri_str;
+		uri_str = gnome_vfs_uri_to_string (uri->parent, hide_options);
 		g_string_prepend_c (string, '#');
-		g_string_prepend (string, gnome_vfs_uri_to_string (uri->parent, 
-								   hide_options));
+		g_string_prepend (string, uri_str);
+		g_free (uri_str);
 	}
 
 	result = string->str;
@@ -1947,7 +1949,8 @@ gnome_vfs_uri_list_free (GList *list)
  * Returns a full URI given a full base URI, and a secondary URI which may
  * be relative.
  *
- * Return value: the URI (NULL for some bad errors).
+ * Return value: a newly allocated string containing the URI 
+ * (NULL for some bad errors).
  **/
 char *
 gnome_vfs_uri_make_full_from_relative (const char *base_uri,
@@ -2037,3 +2040,27 @@ gnome_vfs_uri_list_parse (const gchar* uri_list)
 
 	return g_list_reverse (result);
 }
+
+/* enumerations from "gnome-vfs-uri.h"
+ * This is normally automatically-generated, but glib-mkenums could not
+ * guess the correct get_type() function name for this enum.
+ */
+GType
+gnome_vfs_uri_hide_options_get_type (void)
+{
+  static GType etype = 0;
+  if (etype == 0) {
+    static const GFlagsValue values[] = {
+      { GNOME_VFS_URI_HIDE_NONE, "GNOME_VFS_URI_HIDE_NONE", "none" },
+      { GNOME_VFS_URI_HIDE_USER_NAME, "GNOME_VFS_URI_HIDE_USER_NAME", "user-name" },
+      { GNOME_VFS_URI_HIDE_PASSWORD, "GNOME_VFS_URI_HIDE_PASSWORD", "password" },
+      { GNOME_VFS_URI_HIDE_HOST_NAME, "GNOME_VFS_URI_HIDE_HOST_NAME", "host-name" },
+      { GNOME_VFS_URI_HIDE_HOST_PORT, "GNOME_VFS_URI_HIDE_HOST_PORT", "host-port" },
+      { GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD, "GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD", "toplevel-method" },
+      { GNOME_VFS_URI_HIDE_FRAGMENT_IDENTIFIER, "GNOME_VFS_URI_HIDE_FRAGMENT_IDENTIFIER", "fragment-identifier" },
+      { 0, NULL, NULL }
+    };
+    etype = g_flags_register_static ("GnomeVFSURIHideOptions", values);
+  }
+  return etype;
+}
